package net.fractalkiss.util;
 
/* 
 * Fragger.java
 *
 * Created on June 27, 2002, 8:22 PM
 */

/**
 *
 * @author  Dmitri Pavlenkov
 * @version
 */
import java.util.*;
/** Utility class that performs fragmentation of frames.
 */
public class Fragger extends Util {
  private int levels=8;
  /** Creates new Fragger with 8 default levels
   */
  public Fragger() {
    new Fragger(8);
  }
  /** Creates new Fragger with given levels
   * @param levels number of fragmentation levels
   */
  public Fragger(int levels) {
    this.levels=levels;
  }
  /** Gets a fragmented of frame b from level n at byte position pos
   * @param n fragmentation level
   * @param b frame number
   * @param pos position in frame array
   * @return returns requested byte
   */
  public char frag(int n,int b, int pos) {
    return ((char[][])frags.get(b))[n][pos];
  }
  
  private LinkedList frags=new LinkedList();
  private LinkedList buf=new LinkedList();
  
  /** Stores a frame in the intermal buffer and calls fragAll when buffer is full,
   * removes a frame when buffer overflows.
   * @param frame char array representing frame bytes
   * @return returns true if fragAll was called, false otherwise
   */  
  public boolean nextFrame(char[] frame) {
    buf.addLast(frame);
    if(buf.size()>levels) buf.removeFirst();
    if(buf.size()==levels) {
      fragAll(buf,levels);
      return true;
    }
    return false;
  }
  /** Populates given fractal table from internal frag buffer.
   * @param plane 2D char array
   * @param frame number of bytes per frame
   */  
  public void workPlane(char[][] plane, int frame) {
    for(int d=1;d<levels;d++) {
      for(int pos=0;pos<frame;pos++) {
        char x=frag(d,0,pos);
        char y=frag(d-1,1,pos);
        plane[x][y]|=(1<<(d-1));
        if(d<2) {
          plane[y][x]|=(1<<(levels-2+d));
        }
      }
    }
  }
  /** Populates fractal 3D table from internal frag table
   * @param cube 3D char array
   * @param frame number of bytes per frame
   */  
  public void workCube(char[][][] cube, int frame) {
    for(int d=2;d<levels;d++) {
      for(int pos=0;pos<frame;pos++) {
        char x=frag(d,0,pos);
        char y=frag(d-1,1,pos);
        char z=frag(d-2,2,pos);
        cube[x][y][z]|=(1<<(d-2));
        if(d<4) {
          cube[z][y][x]|=(1<<(levels-4+d));
        }
      }
    }
  }
  
  /** Makes sure all necessary fragments are in memory
   * @param frames bytes per frame
   * @param levels buffer size, 8
   */
  public void fragAll(LinkedList frames, int levels) {
    if(ixor.size()>0) ixor.removeFirst();
    if(frags.size()<1) {
      for(int b=0;b<levels;b++) {
        frags.addLast(new char[levels][]);
        for(int n=0;n<levels-b;n++) {
          ((char[][])frags.get(b))[n]=frag(frames,n,b);
        }
      }
      return;
    }
    frags.removeFirst();
    frags.addLast(new char[levels][]);
    for(int b=0;b<levels;b++) {
      ((char[][])frags.get(b))[levels-1-b]=frag(frames,levels-1-b,b);
    }
  }
  /** FRAG(n,b) implementation<pre>
   * FRAG(n,b) = INAND(n-1,b) NAND FRAME(b)
   * FRAG(0,b) = FRAME(b)
   * </pre>
   * @param frames bytes per frame
   * @param n fragmentation level
   * @param b frame number
   * @return frame b fragmented at level n
   */
  public char[] frag(LinkedList frames, int n, int b) {
    if(n==0) return (char[])((char[])frames.get(b));
    return nand((char[])frames.get(b),inand(frames,n-1,b));
  }
  /** Implementation of INAND(n,b) or iterative nand<pre>
   * INAND(n,b) = IXOR(n,b) NAND IXOR(n-1,b) NAND...NAND IXOR(0,b)
   * INAND(0,b) = IXOR(0,b)
   * iterates nanding xor masks from bottom to top
   * </pre>
   *
   * @param frames bytes per frame
   * @param n nand level
   * @param b frame number
   * @return dividing shape
   */
  public char[] inand(LinkedList frames,int n,int b) {
    if(n==0) return ixor(frames,n,b);
    char[] xor=nand(ixor(frames,n-1,b),ixor(frames,n,b));
    for(int i=n-2;i>-1;i--) {
      xor=nand(ixor(frames,i,b),xor);
    }
    return xor;
  }
  private LinkedList ixor=new LinkedList();
  
  /** convenience method to retreive stored IXOR from LinkedList
   * @param n xor level
   * @param b frame number
   * @return xor mask for every byte
   */
  public char[] ixor(int n,int b) {
    if(ixor==null) return null;
    if(ixor.get(b)==null) return null;
    //if(((char[][])ixor.get(b))[n]!=null)
    //System.out.println("ixor("+n+","+b+") = "+new String(((char[][])ixor.get(b))[n]));
    return ((char[][])ixor.get(b))[n];
  }
  private void ixorPut(int n,int b, char[] xor) {
    ((char[][])ixor.get(b))[n]=xor;
  }
  /** Implementation of IXOR(n,b) or recursive xor
   * <pre>
   * IXOR(n,b) = IXOR(n-1,b) XOR IXOR(n-1,b plus 1) (go figure, plus sign doesn't show)
   * IXOR(0,b) = FRAME(b) XOR FRAME (b plus 1)
   * </pre>
   * calculates xor masks of shapes and their movements
   *
   * @param frames bytes per frame
   * @param n xor level
   * @param b frame number
   * @return xor mask
   */
  public char[] ixor(LinkedList frames,int n,int b) {
    if(ixor.size()<b+1) {
      for(int i=0;i<(b-ixor.size()+1);i++) {
        ixor.addLast(new char[levels][]);
      }
    }
    if(ixor(n,b)!=null) return ixor(n,b);
    if(n==0) {
      ixorPut(0,b,xor((char[])frames.get(b),(char[])frames.get(b+1)));
      return ixor(0,b);
    }
    ixorPut(n,b,xor(ixor(frames,n-1,b),ixor(frames,n-1,b+1)));
    return ixor(n,b);
  }
}
